@@ -3,10 +3,10 @@
 from deployv_addon_gitlab_tools.get_addons import get_module_list
 from deployv.helpers import utils
 import distutils.spawn
-from docker import errors, APIClient as Client
+from docker import errors, from_env
 import json
 import logging
-from os import environ, path, makedirs, chmod, remove
+from os import environ, path, makedirs, chmod, remove, getcwd, chdir, listdir, rmdir
 import re
 import requests
 import shlex
@@ -14,10 +14,11 @@ import subprocess
 import sys
 import time
 from urllib3.exceptions import ReadTimeoutError
+from shutil import copy, rmtree
 
 
 _logger = logging.getLogger('deployv.' + __name__)
-_cli = Client(timeout=7200)
+_cli = from_env(timeout=7200).api
 TO_SCAN = ['git.vauxoo.com',
            'github.com',
            'gitlab.com',
@@ -84,7 +85,7 @@ def run_build_image(config):
     tag = short_id
     if config.get('docker_image_repo'):
         tag = "%s-%s-%s" % (config['main_app'], config['version'], short_id)
-        if config.get('ci_commit_ref_name') != config['version']:
+        if is_dev_repo(config):
             tag = "%s-%s-dev%s" % (config['main_app'], config['version'], short_id)
     config.update({'image_tag': tag})
 
@@ -167,6 +168,10 @@ def get_value(var_name, **kwargs):
     return kwargs[var_name] if var_name in kwargs else environ[var_name.upper()]
 
 
+def is_dev_repo(config):
+    return config.get('ci_commit_ref_name') != config['version'] or config.get('ci_project_namespace').endswith('-dev')
+
+
 def get_config(**kwargs):
     rand_str = 'hd103sl'
     res = {}
@@ -215,6 +220,16 @@ def get_config(**kwargs):
     res.update({
         'main_app': get_main_app(res)
     })
+
+    # Host folder where the artifacts will be stored after the testes were executed
+    res.update({'container_vol': res.get('ci_commit_ref_slug')})
+
+    # The work dir inside the container, the config and coverage result will be stored here
+    res.update({'coverage_workdir': '/home/odoo'})
+
+    # Where the results will be stored, html and xml files. Ths odoo logs will be stored here, but won't be uploaded
+    # to the coverage server
+    res.update({'coverage_reports': '/tmp/coverage_reports'})
     return res
 
 
@@ -265,7 +280,7 @@ def scan_keys(folder):
 
 
 def add_private_key(key, folder):
-    """ Generates the id_rsa file if it doesn't exits with the proper
+    """ Generates the id_rsa file if it doesn't exist with the proper
     format and permissions
 
     :param key: The key content
@@ -334,7 +349,10 @@ def start_instance(config):
         "DB_USER": "odoo",
         "DB_PASSWORD": "odoo",
         "DB_HOST": config['postgres_container'],
-        "ODOO_CONFIG_FILE": "/home/odoo/.openerp_serverrc"
+        "ODOO_CONFIG_FILE": "/home/odoo/.openerp_serverrc",
+        # Enable this to debug coverage
+        # "COVERAGE_DEBUG": "config,dataio,process",
+        "COVERAGE_RCFILE": "/home/odoo/.coveragerc"
     }
     for env_var in ['COUNTRY', 'LANGUAGE']:
         env.update({env_var: environ.get(env_var, "")})
@@ -369,21 +387,23 @@ def start_postgres(config):
     _logger.info(container)
 
 
-def exec_cmd(container, cmd, user=None, stream=False):
+def exec_cmd(container, cmd, user=None, stream=False, workdir=None, tty=False):
     lines = []
     container_id = _cli.inspect_container(container).get('Id')
     _logger.debug('Executing command "{cmd}" in container "{con}".'.format(cmd=cmd, con=container))
     try:
-        exec_id = _cli.exec_create(container_id, cmd, user=user)
+        exec_id = _cli.exec_create(container_id, cmd, user=user, workdir=workdir, tty=tty)
     except errors.APIError as error:
         _logger.error('Error: %s', error.explanation)
         raise
     res = _cli.exec_start(exec_id.get('Id'), stream=stream)
     if stream:
+        _logger.info("--->> Start of the stream <<---")
         for line in res:
             line = utils.decode(line)
-            _logger.info(line.strip('\n'))
+            print(line.strip('\n'))
             lines.append(line)
+        _logger.info("--->> End of the stream <<---")
         return lines
     return utils.decode(res)
 
@@ -407,10 +427,11 @@ def is_running(config):
     retries = 0
     while retry and retries <= 10:
         try:
-            res = exec_cmd(config['instance_image'], 'supervisorctl status odoo')
-        except errors.APIError:
+            res = exec_cmd(config['container_name'], 'supervisorctl status odoo')
+        except errors.APIError as err:
             retries += 1
             _logger.warn('Container error, retrying %s', retries)
+            _logger.debug('Error: %s', err)
             time.sleep(5)
             continue
         _logger.info('is_running: %s', res.strip())
@@ -464,6 +485,133 @@ def install_module(config):
         return False
     return True
 
+def test_module(config):
+    module = config['main_app']
+    extra = ''
+    if config.get('language', False):
+        extra += ' --load-language={lang}'.format(lang=config.get('language'))
+    test_tags = get_test_modules('.')
+    if test_tags != '':
+        extra += ' --test-tags={tags}'.format(tags=test_tags)
+    _logger.debug("test tags: %s", test_tags)
+    test_command = (
+        "unbuffer coverage run /home/odoo/instance/odoo/odoo-bin -d test -i {mod} --test-enable --workers 0 "
+        "{extra} --stop-after-init".format(mod=module, extra=extra)
+    )
+    _logger.info('Verifying supervisorctl')
+    is_running(config)
+    _logger.info('Stopping odoo')
+    exec_cmd(config['container_name'], 'supervisorctl stop odoo')
+    prepare_coverage(config)
+
+    _logger.info('\nRunning tests for %s', module)
+    _logger.debug('Command : %s', test_command)
+    test_res = exec_cmd(config['container_name'], test_command, 'odoo', stream=True, workdir=config.get('coverage_workdir'))
+
+    test_log = resume_log(test_res, config['allow_deprecated'])
+    show_log(test_log[1], 'Installation with demo and tests', config)
+    save_logs(config, test_res)
+
+    # subprocess.call(['docker', 'cp', config.get('container_name')+':/tmp/overage', cover_data])
+
+    cover_reports = [
+        'coverage report -m --show-missing'.format(workdir=config.get('coverage_reports')),
+        'coverage html -d {workdir}'.format(workdir=config.get('coverage_reports')),
+        'coverage xml -o {workdir}/coverage.xml'.format(workdir=config.get('coverage_reports')),
+    ]
+
+    cwd = getcwd()
+    chdir(config.get('container_vol'))
+    for cr in cover_reports:
+        _logger.info('Running command: %s', cr)
+        cmd_res = exec_cmd(config['container_name'], cr, 'odoo', stream=False, workdir=config.get('coverage_workdir'))
+        _logger.info('Command: "%s" exited with:\n%s', cr, cmd_res)
+
+    reports_getters =[
+        'docker cp {container}:{reports} .'.format(container=config.get('container_name'), reports=config.get('coverage_reports')),
+        'rm .coveragerc',
+    ]
+
+    reports_dir = path.basename(config.get('coverage_reports'))
+    for rg in reports_getters:
+        subprocess.call(shlex.split(rg))
+    copy_files(reports_dir, './')
+    rmtree(reports_dir)
+    chdir(cwd)
+
+
+    return test_log[0]
+
+def copy_files(src, dst):
+    for file_name in listdir(src):
+        source = path.join(src, file_name)
+        destination = path.join(dst, file_name)
+        if path.isfile(source):
+            copy(source, destination)
+
+def prepare_coverage(config):
+    if not path.isdir(config.get('container_vol')):
+        makedirs(config.get('container_vol'))
+    fcoveragerc = path.join(config.get('container_vol'), '.coveragerc')
+    with open(fcoveragerc, "w") as f_coveragerc:
+        f_coveragerc.write(get_coverage_content(config.get('ci_project_name')))
+
+    cp_cover_cmd = 'docker cp {coverrc} {container}:{folder}'.format(
+        coverrc=fcoveragerc,
+        container=config.get('container_name'),
+        folder=config.get('coverage_workdir'))
+    subprocess.call(shlex.split(cp_cover_cmd))
+
+    _logger.debug('Changed permissions for coverage %s', exec_cmd(config['container_name'], 'chown odoo:odoo {coveragerc}'.format(coveragerc=path.join(config.get('coverage_workdir'), '.coveragerc'))))
+    _logger.debug('Create workdir %s', exec_cmd(config['container_name'], 'mkdir {report}'.format(report=config.get('coverage_reports')), user='odoo'))
+
+    _logger.debug('Installing dependencies to run tests')
+    exec_cmd(config.get('container_name'), 'touch /home/odoo/full_test-requirements.txt')
+    exec_cmd(config.get('container_name'), 'pip3 install -r /home/odoo/full_test-requirements.txt')
+
+    exec_cmd(config.get('container_name'), 'apt update')
+    exec_cmd(config.get('container_name'), 'apt install expect-dev')
+
+def save_logs(config, logs):
+    odoo_log = path.join(config.get('container_vol'), 'odoo.log')
+    _logger.info("Save log in %s cwd %s", odoo_log, getcwd())
+    with open(odoo_log, "a") as fodoo_log:
+        for line in logs:
+            fodoo_log.write(line)
+
+
+def get_coverage_content(config_main_app):
+    coveragerc_content = """
+# Config file .coveragerc
+
+[report]
+include =
+    /home/odoo/instance/extra_addons/{main_app}/*
+
+omit =
+    */scenario/*
+    */scenarios/*
+    */test/*
+    */tests/*
+    *_example/*
+    *__init__.py
+    *__openerp__.py
+    *__manifest__.py
+    {exclude_var}
+
+# Regexes for lines to exclude from consideration
+exclude_lines =
+    # Have to re-enable the standard pragma
+    pragma: no cover
+    # Don't complain about null context checking
+    if context is None:
+    # Don't check obviously not implemented
+    raise NotImplementedError
+    # We don't really care what happens if fail
+    except ImportError:
+
+    """.format(main_app=config_main_app, workdir='/tmp', exclude_var='${EXCLUDE_COVERAGE}')
+    return coveragerc_content
 
 def show_log(log, title, config):
     _logger.info('='*50)
@@ -513,7 +661,8 @@ def resume_log(log_lines, allow_deprecated=False):
         criteria = re.compile(r'.*\d\sWARNING\s.*')
         return criteria.match(line) \
             and 'no translation for language' not in line \
-            and 'Deprecated method' not in line
+            and 'Deprecated method' not in line \
+            and 'odoo.tests.runner: 0 failed, 0 error(s) of 0 tests when loading database' not in line
 
     filters = {
         'critical': critical,
@@ -525,11 +674,16 @@ def resume_log(log_lines, allow_deprecated=False):
     }
     success = True
     res = {name: [] for name in filters}
-    warn_to_replace = '\x1b[1;33m\x1b[1;49mWARNING\x1b[0m'
+    # Read log file removing ASCII color escapes:
+    # http://serverfault.com/questions/71285
+    color_regex = re.compile(r'\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[m|K]')
+    is_magic_line = False
     for line in log_lines:
         original_line = line.strip()
+        if 'odoo.modules.loading: Modules loaded.' in original_line:
+            is_magic_line = True
         # Remove colorized when exists to regex
-        stripped_line = original_line.replace(warn_to_replace, 'WARNING')
+        stripped_line = color_regex.sub('', original_line)
         for name, criteria in filters.items():
             if criteria(stripped_line):
                 if name in ['critical', 'errors']:
@@ -538,15 +692,22 @@ def resume_log(log_lines, allow_deprecated=False):
                     success = False
                 res.get(name).append(original_line)
                 break
-    return success, res
+    return success and is_magic_line, res
 
 
-def run_image_tests(config):
+def run_image_tests(config, test_only=False):
     clean_containers(config)
     start_postgres(config)
     resp = create_postgres_user(config)
     _logger.info(resp)
     start_instance(config)
-    res = install_module(config)
+    if not test_only:
+        res = install_module(config)
+    else:
+        res = test_module(config)
     clean_containers(config)
     return res
+
+def get_test_modules(folder):
+    return ','.join(['/' + i for i in listdir(folder) if path.isdir(path.join(folder, i)) and path.isfile(path.join(folder, i, '__manifest__.py'))])
+
